/*
 * Copyright © 2006 Carl Worth
 *
 * This program is free software; you can redistribute it and\/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA."
 */

#ifndef _DICT_H_
#define _DICT_H_

/* Portably, schmortability. I want ease of programming. */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

#ifndef FALSE
# define FALSE 0
#endif

#ifndef TRUE
# define TRUE 1
#endif

typedef int bool_t;

typedef struct _trie {
    uint32_t flags;
    struct _trie *next[26];
} trie_t;

typedef trie_t dict_t;
typedef trie_t *dict_cursor_t;
typedef uint32_t dict_entry_t;

/* Initialization and cleanup */
void
dict_init (dict_t *dict);

void
dict_fini (dict_t *dict);

/* Adding new words */
void
dict_add_word (dict_t		*dict,
	       const char	*word);
void
dict_add_words_from_file (dict_t	*dict,
			  const char	*filename);

/* Looking up an entry in the dictionary */
dict_entry_t *
dict_lookup (dict_t	*dict,
	     const char	*word);

typedef bool_t
(*dict_entry_predicate_t) (dict_entry_t entry);

int
dict_count (dict_t			*dict,
	    dict_entry_predicate_t	predicate);

/* Querying a dictionary entry. The dict interface uses 1 bit.
 * All of the remaining bits are available for application use.
 */
#define DICT_ENTRY_FLAG_IS_WORD		(1<<0)

#define DICT_ENTRY_IS_WORD(entry) ((entry) && ((*entry) & DICT_ENTRY_FLAG_IS_WORD))

/* Printing the dictionary */
int
dict_print (dict_t *dict);

int
dict_print_by_length (dict_t *dict);

int
dict_print_if (dict_t			*dict,
	       dict_entry_predicate_t	 predicate);

int
dict_print_by_length_if (dict_t			*dict,
			 dict_entry_predicate_t	 predicate);

/* More general callback-based iteration of all entries */
typedef void (* dict_action_t) (void *closure, char *word, dict_entry_t *entry);

int
dict_for_each (dict_t		*dict,
	       dict_action_t	 action,
	       void		*closure);

int
dict_for_each_by_length (dict_t		*dict,
			 dict_action_t	 action,
			 void		*closure);

int
dict_for_each_if (dict_t			*dict,
		  dict_action_t			 action,
		  void				*closure,
		  dict_entry_predicate_t	 predicate);

int
dict_for_each_by_length_if (dict_t			*dict,
			    dict_action_t		 action,
			    void			*closure,
			    dict_entry_predicate_t	 predicate);

/* Character-by-character perusal of the dictionary */
dict_cursor_t
dict_root (dict_t *dict);

dict_cursor_t
dict_cursor_next (dict_cursor_t cursor,
		  char		next);

dict_entry_t *
dict_cursor_resolve (dict_cursor_t cursor);

#define DICT_CURSOR_NIL NULL

#endif /* _DICT_H_ */
