/*
 * Copyright © 2006 Carl Worth
 *
 * This program is free software; you can redistribute it and\/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA."
 */
#include <stdlib.h>
#include <string.h>
#include <goocanvas.h>
#include <sys/time.h>
#include <time.h>
#include <ctype.h>
#include <math.h>
#include <gdk/gdkkeysyms.h>

#include "word-game.h"
#include "demo-item.h"

#define RACK_DICT_ENTRY_FOUND (1<<1)
#define MAX_TILES 7

typedef struct _tile
{
    char letter;
    int rack_index;
    int x, y;
    GooCanvasItem *item;
    gboolean guessed;
} tile_t;

typedef struct _rack
{
    tile_t *tiles[MAX_TILES];
    int num_tiles;
    char guess[MAX_TILES+1];
    int guess_length;
    dict_t *solution;
    GooCanvasItem *solution_item;
} rack_t;

#define LETTER_SIZE 60
#define LETTER_PAD 5

static void
guess_tile_position (int i, int *x, int *y)
{
    *x = 20 + i * (LETTER_SIZE + LETTER_PAD);
    *y = LETTER_PAD;
}

static void
rack_tile_position (int i, int *x, int *y)
{
    guess_tile_position (i, x, y);
    *y += (LETTER_SIZE + LETTER_PAD);
}

typedef struct _dict_paint_cursor
{
    cairo_t *cr;
    int line_height;
    int x;
    int y;
    int max_column_width;
    int max_y;
} dict_paint_cursor_t;

static void
dict_paint_action (void *closure, char *word, dict_entry_t *entry)
{
    dict_paint_cursor_t *cursor = closure;
    cairo_t *cr = cursor->cr;
    double new_x, new_y;

    if (strlen (word) < 3)
	return;

    cairo_move_to (cr, cursor->x, cursor->y);
    if (*entry & RACK_DICT_ENTRY_FOUND) {
	cairo_show_text (cr, word);
    } else {
	int i, length = strlen (word);
	for (i = 0; i < length; i++)
	    cairo_show_text (cr, "_");
    }
    cairo_get_current_point (cr, &new_x, &new_y);
    if (new_x > cursor->max_column_width)
	cursor->max_column_width = new_x;
    cursor->y += cursor->line_height;
    if (cursor->y > cursor->max_y) {
	cursor->x = cursor->max_column_width + cursor->line_height / 2;
	cursor->y = 0;
    }
}

#define SOLUTION_FONT_SIZE	12
#define SOLUTION_LINE_HEIGHT	(1.5 * SOLUTION_FONT_SIZE)

static void
dict_paint (cairo_t *cr, void *closure, double width, double height)
{
    dict_t *dict = closure;
    dict_paint_cursor_t cursor;

    cairo_save (cr);
    cairo_set_source_rgb (cr, 0.0, 0.0, 0.0); /* black */

    cursor.cr = cr;

    cairo_select_font_face (cr, "mono", 0, 0);
    cairo_set_font_size (cr, SOLUTION_FONT_SIZE);
    cursor.line_height = SOLUTION_LINE_HEIGHT;
    
    cursor.x = 0;
    cursor.y = 0;

    cursor.max_column_width = 0;
    cursor.max_y = height;

    dict_for_each_by_length (dict,
			     dict_paint_action,
			     &cursor);

    cairo_restore (cr);
}

static void
tile_paint (cairo_t *cr, void *closure, double width, double height)
{
    tile_t *tile = closure;

    cairo_pattern_t *gradient;
    cairo_text_extents_t extents;
    int rad = (int) MIN (width / 2, height / 2);
    int cx = width / 2;
    int cy = cx;
    int tx, ty;
    double spot_angle = M_PI / 4.0;
    double spot_rad = rad / 2.0;
    char string[2];

    cairo_save (cr);

    gradient = cairo_pattern_create_radial (cx - spot_rad * cos (spot_angle),
					    cy - spot_rad * sin (spot_angle),
					    0.0,
					    cx - spot_rad * cos (spot_angle),
					    cy - spot_rad * sin (spot_angle),
					    rad + spot_rad);
    cairo_pattern_add_color_stop_rgb (gradient, 0.0, 1.0, 1.0, 1.0);
    cairo_pattern_add_color_stop_rgb (gradient, 1.0, 0.33, 0.33, 0.33);

    cairo_set_source (cr, gradient);

    cairo_arc (cr,
	       cx, cy,
	       rad, 0, 2 * M_PI);

    cairo_fill (cr);

    cairo_select_font_face (cr, "mono",
			    CAIRO_FONT_SLANT_NORMAL,
			    CAIRO_FONT_WEIGHT_BOLD);
    cairo_set_font_size (cr, 1.8 * rad);

    string[0] = tile->letter;
    string[1] = '\0';
    cairo_text_extents (cr, string, &extents);
    tx = cx - extents.width / 2 - extents.x_bearing;
    ty = cy - extents.height / 2 - extents.y_bearing;

    cairo_set_source_rgb (cr, 0.7, 0.7, 0.7);
    cairo_move_to (cr, tx + 1, ty + 1);
    cairo_show_text (cr, string);
		
    cairo_set_source_rgb (cr, 0.33, 0.33, 0.33);
    cairo_move_to (cr, tx - 1, ty - 1);
    cairo_show_text (cr, string);

    cairo_set_source_rgb (cr, 0.2, 0.3, 0.8);
    cairo_move_to (cr, tx, ty);
    cairo_show_text (cr, string);

    cairo_restore (cr);
}

static void
tile_glide_to (tile_t *tile, int x, int y)
{
    goo_canvas_item_animate (tile->item, x, y,
			     1.0, 0,
			     500, 40,
			     GOO_CANVAS_ANIMATE_FREEZE);
    tile->x = x;
    tile->y = y;
}

static tile_t *
tile_create (GooCanvasItem *parent,
	     char letter, int rack_index)
{
    tile_t *tile;

    tile = g_malloc (sizeof (tile_t));
    tile->letter = tolower (letter);
    tile->rack_index = rack_index;
    rack_tile_position (rack_index, &tile->x, &tile->y);
    tile->item = goo_demo_item_new (parent,
				    tile->x, tile->y,
				    LETTER_SIZE, LETTER_SIZE,
				    tile_paint,
				    tile, NULL);

    tile->guessed = FALSE;

    return tile;
}

static gboolean
on_delete_event (GtkWidget *window,
		 GdkEvent  *event,
		 gpointer   unused_data)
{
    exit (0);
}

static int
rand_within (int num_values)
{
    return (int) ((double) num_values * (rand() / (RAND_MAX + 1.0)));
}

static void
shuffle (int *array, int length)
{
    int i, r, tmp;

    for (i = 0; i < length; i++)
    {
	r = i + rand_within (length - i);
	tmp = array[i];
	array[i] = array[r];
	array[r] = tmp;
    }
}

static void
rack_init (rack_t	*rack,
	   dict_t	*solution)
{
    int i;

    for (i = 0; i < MAX_TILES; i++)
	rack->tiles[i] = NULL;
    rack->num_tiles = 0;

    rack->guess_length = 0;
    rack->solution = solution;
    rack->solution_item = NULL;
}

static void
rack_create_tiles (rack_t	*rack,
		   GooCanvas	*canvas,
		   char		*word)
{
    int i;

    GooCanvasItem *root = goo_canvas_get_root_item (canvas);

    for (i = 0; i < MIN (MAX_TILES, strlen (word)); i++)
	rack->tiles[i] = tile_create (root, word[i], i);
    rack->num_tiles = i;
}

static gboolean
rack_shuffle (rack_t *rack)
{
    int indices[MAX_TILES];
    int i, x, y;

    for (i = 0; i < rack->num_tiles; i++)
	indices[i] = i;

    shuffle (indices, rack->num_tiles);

    for (i = 0; i < rack->num_tiles; i++) {
	rack->tiles[i]->rack_index = indices[i];
	rack_tile_position (indices[i], &x, &y);
	tile_glide_to (rack->tiles[i], x, y);
    }

    return TRUE;
}

static void
rack_return_tile (rack_t *rack, tile_t *tile)
{
    int x, y;

    rack_tile_position (tile->rack_index, &x, &y);
    tile_glide_to (tile, x, y);
    tile->guessed = FALSE;
    rack->guess_length--;
    rack->guess[rack->guess_length] = '\0';
}

static void
rack_return_all (rack_t *rack)
{
    int i;

    for (i = 0; i < rack->num_tiles; i++) {
	if (rack->tiles[i]->guessed)
	    rack_return_tile (rack, rack->tiles[i]);
    }
    rack->guess_length = 0;
    rack->guess[0] = '\0';
}

static gboolean
on_key_press_event (GtkWidget	*widget,
		    GdkEventKey	*event,
		    gpointer	 user_data)
{
    int i, x, y;
    char guess_letter;
    rack_t *rack = user_data;

    if (event->keyval == GDK_Return) {
	dict_entry_t *entry;
	if (rack->guess_length >= 3) {
	    entry = dict_lookup (rack->solution, rack->guess);
	    if (DICT_ENTRY_IS_WORD (entry)) {
		*entry = *entry | RACK_DICT_ENTRY_FOUND;
		goo_canvas_item_simple_changed (GOO_CANVAS_ITEM_SIMPLE (rack->solution_item), FALSE);
	    }
	}
	rack_return_all (rack);
	return TRUE;
    }

    if (event->keyval == GDK_space) {
	rack_return_all (rack);
	rack_shuffle (rack);
	return TRUE;
    }

    if (event->keyval == GDK_BackSpace) {
	gboolean found = FALSE;
	int found_index;
	x = 0;
	for (i = 0; i < rack->num_tiles; i++) {
	    /* XXX: evil stuff here... please refactor a lot */
	    if (rack->guess[rack->guess_length-1] == rack->tiles[i]->letter &&
		rack->tiles[i]->guessed &&
		rack->tiles[i]->x > x)
	    {
		found = TRUE;
		found_index = i;
	    }
	}
	if (found) {
	    rack_return_tile (rack, rack->tiles[found_index]);
	    return TRUE;
	}
	return FALSE;
    }

    /* XXX: event->string is deprecated, but the non-deprecated
     * input-method stuff (GtkIMContext) is extremely non-obvious to
     * use. */
    guess_letter = tolower (event->string[0]);
    for (i = 0; i < rack->num_tiles; i++) {
	if (guess_letter == rack->tiles[i]->letter && 
	    ! rack->tiles[i]->guessed)
	{
	    guess_tile_position (rack->guess_length, &x, &y);
	    tile_glide_to (rack->tiles[i], x, y);
	    rack->tiles[i]->guessed = TRUE;
	    rack->guess[rack->guess_length++] = guess_letter;
	    rack->guess[rack->guess_length] = '\0';
	    return TRUE;
	}
    }

    return FALSE;
}

static GtkWidget *
create_window (rack_t *rack)
{
    GtkWidget *window, *scrolled_window;

    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_default_size (GTK_WINDOW (window), 500, 500);
    gtk_widget_show (window);
    g_signal_connect (window, "delete_event",
		      (GtkSignalFunc) on_delete_event, NULL);

    gtk_widget_add_events (window, GDK_KEY_PRESS_MASK);
    g_signal_connect (window, "key_press_event",
		      (GtkSignalFunc) on_key_press_event, rack);

    scrolled_window = gtk_scrolled_window_new (NULL, NULL);
    gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolled_window),
					 GTK_SHADOW_IN);
    gtk_scrolled_window_set_policy  (GTK_SCROLLED_WINDOW (scrolled_window),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_AUTOMATIC);
    gtk_widget_show (scrolled_window);
    gtk_container_add (GTK_CONTAINER (window), scrolled_window);

    return scrolled_window;
}

static GooCanvas *
create_canvas (GtkWidget *parent, rack_t *rack)
{
    GtkWidget *canvas;
    GooCanvasItem *root;

    canvas = goo_canvas_new ();
    gtk_widget_set_size_request (canvas, 400, 400);
    goo_canvas_set_bounds (GOO_CANVAS (canvas), 0, 0, 400, 400);
    gtk_widget_show (canvas);
    gtk_container_add (GTK_CONTAINER (parent), canvas);

    root = goo_canvas_get_root_item (GOO_CANVAS (canvas));

    rack->solution_item = goo_demo_item_new (root,
					     LETTER_PAD + SOLUTION_LINE_HEIGHT,
					     LETTER_PAD + 2 * (LETTER_SIZE + 2 * LETTER_PAD),
					     400, 400 - (2 * (LETTER_SIZE + 2 * LETTER_PAD)),
					     dict_paint, rack->solution,
					     NULL);

    return GOO_CANVAS (canvas);
}

int
main (int argc, char *argv[])
{
    dict_t dict, solution;
    struct timeval tv;
    bag_t bag;
    rack_t rack;
    char word[8];
    int i;
    char *draw;
    GtkWidget *window;
    GooCanvas *canvas;

    gettimeofday (&tv, NULL);
    srand (tv.tv_sec ^ tv.tv_usec);

    bag_init (&bag);
    bag_shuffle (&bag);

    /* Keep drawing until we get 7 non-blank tiles */
    i = 0;
    draw = bag.tiles;
    while (i < 7) {
	if (*draw != '?')
	    word[i++] = toupper (*draw);
	draw++;
    }

    dict_init (&dict);
    dict_add_words_from_file (&dict, "words.txt");

    dict_init (&solution);
    subanagram_expand (word, &dict, &solution);

    gtk_init (&argc, &argv);

    rack_init (&rack, &solution);

    window = create_window (&rack);

    canvas = create_canvas (window, &rack);

    rack_create_tiles (&rack, canvas, word);

    gtk_main ();

    return 0;
}
